"""
Timer which saves history of runs.
"""

import logging
import time
from contextlib import contextmanager


@contextmanager
def timer(name: str, disable: bool = False):
    """Simple timer as context manager."""

    start = time.time()
    yield
    if not disable:
        logging.info(f'[{name}] done in {(time.time() - start)*1000:.1f} ms')


class TimerWithHistory:
    """Timer as context mamager which saves history.

    This timer should be initialized and then used as context manager.
    After each run it appends execution time to list with history.
    Different runs could have different names and history is saved as dict
    with separate key for each name.

    Args:
        default_name (str): Default name for given run.
        disable (bool): If True then disable timer.

    Example:
        timer = TimerWithHistory()
        with timer():
            time.sleep(1)
        with timer(name='first'):
            time.sleep(2)
        # get history
        hist1, hist2 = timer.history['default'], timer.history['first']
    """

    def __init__(self, default_name: str = 'default', disable: bool = False):

        self._start = None
        self.history = {}
        self.name = default_name
        self.default_name = default_name
        self.disable = disable

    def start(self):
        """Start timer."""

        if self._start is not None:
            raise RuntimeError('Timer already started...')
        self._start = time.perf_counter()

    def stop(self):
        """Stop timer and save result to history."""

        if self._start is None:
            raise RuntimeError('Timer not yet started...')
        elapsed = time.perf_counter() - self._start
        if self.history.get(self.name):
            self.history[self.name].append(elapsed)
        else:
            self.history[self.name] = [elapsed]
        self._start = None

    def __enter__(self):
        if not self.disable:
            self.start()
        return self

    def __exit__(self, *args):
        if not self.disable:
            self.stop()

    def __call__(self, name=None):
        if not self.disable:
            self.name = name or self.default_name
        return self



"""
Wrapper for approximate nearest neighbors search using hnswlib library.
"""

from typing import Optional

import hnswlib
import numpy as np


class HNSWIndex:
    """
    Approximate nearest neighbors search using hnswlib library.

    Args:
        max_elements: Maximum number of elements that can be stored in index (hnswlib parameter).
        path: Path to previously saved index. If not None, load it. If None, initialize empty index.
        space: Distance metric (hnswlib parameter). Possible values:
            'l2', 'ip' (inner product), 'cosine.
        dim: Dimensionality of vectors in index (hnswlib parameter).
        ef: hnswlib parameter, defines query time accuracy/speed trade-off.
        ef_construction: hnswlib parameter, defines construction time/accuracy trade-off.
        M: hnswlib parameter, defines maximum number of outgoing connections in the graph.

    Attributes:
        index: Instance of hnswlib.Index.
    """

    def __init__(self, max_elements: int, path: Optional[str] = None,
                 space: str = 'ip', dim: int = 512, ef: int = 100,
                 ef_construction: int = 250, M: int = 16):

        self.index = hnswlib.Index(space=space, dim=dim)

        if path is not None:
            self.index.load_index(path, max_elements=max_elements)
        else:
            self.index.init_index(max_elements=max_elements, ef_construction=ef_construction, M=M)

        self.index.set_ef(ef)

    def add_item(self, vector: np.ndarray, label: Optional[int] = None):
        """Add one element to index.

        Args:
            vector: Numpy array with vector for one element.
            label: Optional integer label for this element.
        """

        self.index.add_items(vector, ids=label)

    def add_items(self, vectors: np.ndarray, labels: Optional[int] = None):
        """Add batch of elements to index.

        Args:
            vectors: Numpy array with vectors for given elements.
            label: Optional integer labels for this elements.
        """

        self.index.add_items(vectors, ids=labels)

    def find_nearest_point(self, vector: np.ndarray):
        """Find nearest point from index for given vector.

        Args:
            vector: Numpy array.

        Returns:
            Label of nearest point and distance to it.
        """

        labels, distances = self.index.knn_query(vector, k=1)
        return labels[0, 0], distances[0, 0]

    def find_nearest_point_batch(self, vectors: np.ndarray):
        """Find nearest point from index for batch of vectors.

        Args:
            vectors: Numpy array.

        Returns:
            Labels of nearest points and corresponding distances to it.
        """

        labels, distances = self.index.knn_query(vectors, k=1)
        return labels[:, 0], distances[:, 0]

    def save(self, path: str):
        """Save index to disk.

        Args:
            path: Save index to this path.
        """

        self.index.save_index(path)



"""
Simple greedy algorithm for clustering with maximum distance between points inside clusters.
"""

from typing import Union

import numpy as np
import numpy_groupies as npg
from scipy.sparse import csr_matrix
from tqdm import tqdm

from .mixins import FitPredictMixin, DistanceMatrixMixin
from .timer import TimerWithHistory


class MaxDiameterClustering(FitPredictMixin, DistanceMatrixMixin):
    """Clustering with maximum diameter (maximum distance between points) inside clusters.

    Args:
        max_distance (float): Maximum distance between points in clusters.
        criterion (str): Criterion for choosing cluster from several candidates.
            If 'distance' then choose cluster with minimum average distance to given point.
            If 'size' then choose cluster with maximum current size.
        metric (str): Distance metric.
            For sparse_dist=True possible options are in sklearn.neighbors.VALID_METRICS['brute'].
            For sparse_dist=False possible options are 'inner_product' or one of metrics
            available in scipy.spatial.distance.pdist. If 'inner_product' then use np.inner
            which is much faster than pdist. 'inner_product' could be used instead
            of cosine distance for normalized vectors.
        precomputed_dist (bool): If True, then input should be precomputed distance matrix,
            if False then input is array with features.
        sparse_dist (bool): If True, then use distance matrix in sparse format (zero elements
            are elements for which distance between points is greater than max_distance).
            If False, then distance matrix is ordinary numpy array.
        deterministic (bool): If True then take points one by one to get determenistic behavior.
            If False then select points at random, so results would be different for each run.
        use_timer (bool): If True then use TimerWithHistory in fit method, which can be accessed
            via self.timer. Can be useful for debugging.
        verbose (bool): If True then output progress info, otherwise be silent.

    Attributes:
        labels_ (np.array): Array with cluster labels after fitting model.
        n_clusters_ (int): Number of clusters after fitting model.
        timer: Timer with history of execution time (access history via self.timer.history).
    """

    def __init__(self, max_distance: float = 0.2, criterion: str = 'distance',
                 metric: str = 'cosine', precomputed_dist: bool = False,
                 sparse_dist: bool = True, deterministic: bool = False,
                 use_timer: bool = False, verbose: bool = True):

        if criterion not in ['size', 'distance']:
            raise ValueError('Wrong criterion value, should be "size" or "distance".')

        self.max_distance = max_distance
        self.criterion = criterion
        self.metric = metric
        self.precomputed_dist = precomputed_dist
        self.sparse_dist = sparse_dist
        self.deterministic = deterministic
        self.use_timer = use_timer
        self.verbose = verbose

        self.labels_ = None
        self.n_clusters_ = None
        self.timer = None

    def fit(self, X: Union[np.ndarray, csr_matrix]):
        """Fit clustering from features or distance matrix.

        Args:
            X (np.ndarray or scipy.sparse.csr_matrix): Array with features or
                precomputed distance matrix, could be in sparse format.
        """

        dist_matrix = self._prepare_distance_matrix(X)

        # create array for labels
        labels = np.empty(dist_matrix.shape[0])
        labels.fill(np.nan)

        # handle case when empty input data is passed
        if len(labels) == 0:
            self.labels_ = labels
            self.n_clusters_ = 0
            return

        # choose first point and assign label to it
        idx = 0 if self.deterministic else np.random.choice(range(len(labels)))
        labels[idx] = 0
        next_cluster = 1

        self.timer = TimerWithHistory(disable=not self.use_timer)

        for _ in tqdm(range(len(labels)-1), desc='MaxDiameterClustering fit',
                      disable=not self.verbose):

            # choose next point
            with self.timer(name='choose_next_point'):
                indexes = np.where(np.isnan(labels))[0]
                idx = indexes[0] if self.deterministic else np.random.choice(indexes)
            # find indices of already labeled points
            with self.timer(name='find_labeled_points'):
                current_cluster_idx = np.where(~np.isnan(labels))[0]
                current_cluster_labels = labels[current_cluster_idx].astype(int)
            # find distances to already labeled points
            with self.timer(name='get_distances'):
                current_dist = self._slice_distance_matrix(dist_matrix, idx, current_cluster_idx)

            # find max distance to each existent cluster
            with self.timer(name='max_distance_to_clusters'):
                cluster_dist_max = npg.aggregate(current_cluster_labels, current_dist,
                                                 func='max', fill_value=np.inf)

            if np.min(cluster_dist_max) <= self.max_distance:
                # find existent clusters with max dist < threshold
                with self.timer(name='candidate_clusters'):
                    candidate_clusters = np.where(cluster_dist_max <= self.max_distance)[0]
                    # directly get label if there is only one such cluster
                    if len(candidate_clusters) == 1:
                        labels[idx] = candidate_clusters[0]
                        continue
                    # otherwise we need to choose between candidate clusters
                    candidate_clusters_idx = np.isin(current_cluster_labels, candidate_clusters)
                    candidate_clusters_labels = current_cluster_labels[candidate_clusters_idx]

                if self.criterion == 'distance':
                    candidate_clusters_dist = current_dist[candidate_clusters_idx]
                    labels[idx] = self._best_candidate_distance(candidate_clusters_labels,
                                                                candidate_clusters_dist)
                elif self.criterion == 'size':
                    labels[idx] = self._best_candidate_size(candidate_clusters_labels)
            else:
                # assign new cluster label
                with self.timer(name='assign_new_label'):
                    labels[idx] = next_cluster
                    next_cluster += 1

        self.labels_ = labels.astype(int)
        self.n_clusters_ = labels.max() + 1

    def _best_candidate_distance(self, candidate_clusters_labels: np.ndarray,
                                 candidate_clusters_dist: np.ndarray) -> int:
        """Find best candidate cluster based on average distance to clusters."""

        # find average distance to clusters
        with self.timer(name='average_distance_to_clusters'):
            cluster_dist_mean = npg.aggregate(candidate_clusters_labels,
                                              candidate_clusters_dist,
                                              func='mean', fill_value=np.inf)

        # assign cluster with min average distance as label
        with self.timer(name='distance_argmin'):
            label = cluster_dist_mean.argmin()

        return label

    def _best_candidate_size(self, candidate_clusters_labels: np.ndarray) -> int:
        """Find best candidate cluster based on size of clusters."""

        # find size of clusters
        with self.timer(name='average_size_of_clusters'):
            cluster_size = npg.aggregate(candidate_clusters_labels,
                                         candidate_clusters_labels,
                                         func='count')
        # assign cluster with max size as label
        with self.timer(name='size_argmax'):
            label = cluster_size.argmax()

        return label



"""
Computation of distance matrix.
"""

import numpy as np
from scipy.sparse import csr_matrix
from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import RadiusNeighborsTransformer


def compute_dist_matrix(X: np.ndarray, metric: str = 'inner_product',
                        fill_diagonal: bool = False) -> np.ndarray:
    """
    Compute distance matrix between points and optionally fill diagonal elements
    with np.inf (may be convenient in some situation).

    Args:
        X (np.ndarray): 2D array with data points.
        metric (str): Distance metric. Possible options are 'inner_product' or one of metrics
            available in scipy.spatial.distance.pdist. If 'inner_product' then use np.inner
            which is much faster than pdist. 'inner_product' could be used instead of cosine
            distance for normalized vectors.
        fill_diagonal (bool): If True then fill diagonal with np.inf.

    Returns:
        Array with shape (len(X), len(X)).
    """

    if X.ndim == 1:
        X = X[None, :]  # for correct work of distance computation

    if metric == 'inner_product':
        dist_matrix = 1 - np.inner(X, X)
    else:
        dist_matrix = pdist(X, metric=metric)
        # squareform converts emmpty dist_matrix array([]) to array([[0.]])
        # this behavior could break the code later
        dist_matrix = squareform(dist_matrix) if len(dist_matrix) > 0 else np.empty((0, 0))

    if fill_diagonal:
        np.fill_diagonal(dist_matrix, np.inf)

    return dist_matrix


def compute_sparse_dist_matrix(X: np.ndarray, metric: str = 'cosine',
                               max_distance: float = 0.2) -> csr_matrix:
    """
    Compute distance matrix in sparse csr format using sklearn RadiusNeighborsTransformer.
    Zero elements of matrix are elements for which distance is greater than max_distance.

    Args:
        X (np.ndarray): 2D array with data points.
        metric (str): Distance metric
            (possible options in sklearn.neighbors.VALID_METRICS['brute']).
        max_distance (float): Maximum distance threshold.

    Returns:
        scipy.sparse.csr_matrix with shape (len(X), len(X)).
    """

    transformer = RadiusNeighborsTransformer(mode='distance', algorithm='brute',
                                             metric=metric, radius=max_distance)

    return transformer.fit_transform(X)



"""
Version of Leader clustering using approximate nearest neighbors search.
"""

import numpy as np
from tqdm import tqdm

from .hnsw import HNSWIndex
from ..timer import timer


class ApproxLeaderClustering:
    """Leader clustering algorithm with approximate nearest neighbors search.

    Approximate nearest neighbors index is used to store leaders of clusters
    and to find nearest leader for new points.

    Args:
        ann_index: instance of HNSWIndex.
        max_radius: Maximum radius of each cluster
            (maximum distance between the leader and all other points in cluster).
        deterministic: If True then take points one by one to get deterministic behavior.
            If False then select points at random, so results would be different for each run.
        verbose: If True then output progress info, otherwise be silent.

    Attributes:
        labels_ (np.ndarray): Array with cluster labels after fitting model.
        n_clusters_ (int): Number of clusters after fitting model.
        centers_ (np.ndarray): Array with indexes of cluster centers.
        leaders_ (np.ndarray): Array with 1 for cluster centers and with 0 for all other points.

    Examples:
        import numpy as np
        from diameter_clustering.approx import HNSWIndex
        from diameter_clustering.approx import ApproxLeaderClustering

        # fit model
        data = np.random.rand(1000, 50)
        hnsw_index = HNSWIndex(max_elements=len(data), space='cosine', dim=50,
                               ef=100, ef_construction=200, M=16)
        model = ApproxLeaderClustering(hnsw_index, max_radius=0.2, deterministic=True)
        labels = model.fit_predict(data)

        # save index for later usage
        hnsw_index.save('hnsw_index.bin')

        # predict clusters for new data later
        new_data = np.random.rand(100, 50)
        hnsw_index = HNSWIndex(max_elements=len(new_data), path='hnsw_index.bin',
                               space='cosine', dim=50, ef=100)
        model = ApproxLeaderClustering(hnsw_index, max_radius=0.2, deterministic=True)
        new_labels = model.predict(new_data)
    """

    def __init__(self, ann_index: HNSWIndex, max_radius: float = 0.1,
                 deterministic: bool = True, verbose: bool = True):

        self.ann_index = ann_index
        self.max_radius = max_radius
        self.deterministic = deterministic
        self.verbose = verbose

        self.labels_ = None
        self.n_clusters_ = None
        self.centers_ = None
        self.leaders_ = None

    def fit(self, X: np.ndarray):
        """Fit clustering.

        Args:
            X: Array with features.
        """

        # create arrays for labels and leaders
        labels = np.empty(len(X))
        labels.fill(np.nan)
        leaders = np.zeros(len(X))
        centers = []

        # handle case when empty input data is passed
        if len(labels) == 0:
            self.labels_ = labels
            self.leaders_ = leaders
            self.n_clusters_ = 0
            return

        # choose first point and assign label to it
        idx = 0 if self.deterministic else np.random.choice(range(len(labels)))
        labels[idx] = 0
        next_cluster = 1
        leaders[idx] = 1
        centers.append(idx)
        self.ann_index.add_item(X[idx])

        for _ in tqdm(range(len(labels)-1), desc='ApproxLeaderClustering fit',
                      disable=not self.verbose):

            # choose next point
            indexes = np.where(np.isnan(labels))[0]
            idx = indexes[0] if self.deterministic else np.random.choice(indexes)

            # find nearest leader
            nearest_leader_idx, nearest_leader_dist = self.ann_index.find_nearest_point(X[idx])

            if nearest_leader_dist <= self.max_radius:
                # assign cluster with nearest leader as label
                labels[idx] = nearest_leader_idx
            else:
                # assign new cluster label
                labels[idx] = next_cluster
                leaders[idx] = 1
                centers.append(idx)
                next_cluster += 1
                self.ann_index.add_item(X[idx])

        self.labels_ = labels.astype(int)
        self.n_clusters_ = int(labels.max() + 1)
        self.centers_ = np.array(centers)
        self.leaders_ = leaders.astype(int)

    def fit_predict(self, X: np.ndarray) -> np.ndarray:
        """Fit clustering and return cluster labels.

        Args:
            X: Array with features.

        Returns:
            Numpy array with labels for data points in X.
        """

        self.fit(X)

        return self.labels_

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Assigning new points to existent clusters without making new clusters.

        Returning -1 for points which can't be assigned to any cluster.
        Finding nearest leaders for points one by one.

        Args:
            X: Array with features for new points.

        Returns:
            Numpy array with labels for data points in X.
        """

        # create array for new labels
        labels = np.empty(len(X))

        for idx in tqdm(range(len(X)), desc='ApproxLeaderClustering assign points to clusters',
                        disable=not self.verbose):

            # find nearest leader
            nearest_leader_idx, nearest_leader_dist = self.ann_index.find_nearest_point(X[idx])

            if nearest_leader_dist <= self.max_radius:
                # assign cluster with nearest leader as label
                labels[idx] = nearest_leader_idx
            else:
                # assign -1 for point which is not close enough to any leader
                labels[idx] = -1

        return labels


    def predict_batch(self, X: np.ndarray) -> np.ndarray:
        """Assigning new points to existent clusters without making new clusters.

        Returning -1 for points which can't be assigned to any cluster.
        Finding nearest leaders for all points at once.

        Args:
            X: Array with features for new points.

        Returns:
            Numpy array with labels for data points in X.
        """

        # create array for new labels
        labels = np.empty(len(X))

        # find nearest leaders
        with timer('find_nearest_point_batch', disable=not self.verbose):
            nearest_leaders_idx, nearest_leaders_dist = \
                self.ann_index.find_nearest_point_batch(X)

        for idx in range(len(X)):

            if nearest_leaders_dist[idx] <= self.max_radius:
                # assign cluster with nearest leader as label
                labels[idx] = nearest_leaders_idx[idx]
            else:
                # assign -1 for point which is not close enough to any leader
                labels[idx] = -1

        return labels



"""
Mixins for clustering algorithms.
"""

import logging
from typing import Union

import numpy as np
from scipy.sparse import csr_matrix

from .dist_matrix import compute_dist_matrix, compute_sparse_dist_matrix
from .timer import timer


class FitPredictMixin:
    """Mixin with fit_predict method."""

    def fit_predict(self, X: Union[np.ndarray, csr_matrix]) -> np.ndarray:
        """Fit clustering from features or distance matrix and return cluster labels.

        Args:
            X (np.ndarray or scipy.sparse.csr_matrix): Array with features
                or precomputed distance matrix, which could be in sparse matrix format.

        Returns:
            Array with cluster labels.
        """

        self.fit(X)

        return self.labels_


class DistanceMatrixMixin:
    """Mixin with methods for working with distance matrix."""

    def _prepare_distance_matrix(self, X: Union[np.ndarray, csr_matrix]):
        """Prepare distance matrix.

        If self.precomputed_dist is True then do nothing, only check for correctness of X.
        Otherwise compute distance matrix regarding X as array of features. If self.sparse_dist
        is True then compute matrix in sparse format."""

        if not self.precomputed_dist:
            if self.sparse_dist:
                if self.verbose:
                    logging.info('computing distance matrix in sparse format...')
                with timer('compute_sparse_dist_matrix', disable=not self.verbose):
                    return compute_sparse_dist_matrix(X, metric=self.metric,
                                                      max_distance=self.max_distance)
            else:
                if self.verbose:
                    logging.info('computing distance matrix in dense format...')
                with timer('compute_dist_matrix', disable=not self.verbose):
                    return compute_dist_matrix(X, metric=self.metric)

        if X.shape[0] != X.shape[1]:
            raise ValueError(f'Distance matrix should be square. Got matrix of shape {X.shape}.')

        if self.sparse_dist:
            if not isinstance(X, csr_matrix):
                raise TypeError('Sparse distance matrix should be in '
                                'scipy.sparse.csr_matrix format.')
        elif not isinstance(X, np.ndarray):
            raise TypeError('Dense distance matrix should be '
                            'an instance of np.ndarray.')

        return X

    def _slice_distance_matrix(self, dist_matrix: Union[np.ndarray, csr_matrix],
                               idx: int, indexes: np.ndarray):
        """Get one row of distance matrix.
        Get distance between given point and several other points.

        Args:
            dist (np.ndarray or scipy.sparse.csr_matrix): Distance matrix.
            idx (int): Index of given point.
            indexes (np.ndarray): Indexes of other points.
        """

        if isinstance(dist_matrix, csr_matrix):
            current_dist = dist_matrix[idx, indexes].toarray()[0, :]
            current_dist[current_dist == 0] = np.inf
        else:
            current_dist = dist_matrix[idx, indexes]

        return current_dist



"""
Implementation of Leader clustering.
"""

from typing import Union

import numpy as np
from scipy.sparse import csr_matrix
from tqdm import tqdm

from .mixins import FitPredictMixin, DistanceMatrixMixin


class LeaderClustering(FitPredictMixin, DistanceMatrixMixin):
    """Leader clustering algorithm.

    Args:
        max_radius (float): Maximum radius of cluster
            (maximum distance between leader and all other points in cluster).
        change_leaders (bool): if True then change cluster leader if there is a point with smaller
            average distance to all points in cluster.
        metric (str): Distance metric.
            For sparse_dist=True possible options are in sklearn.neighbors.VALID_METRICS['brute'].
            For sparse_dist=False possible options are 'inner_product' or one of metrics
            available in scipy.spatial.distance.pdist. If 'inner_product' then use np.inner
            which is much faster than pdist. 'inner_product' could be used instead
            of cosine distance for normalized vectors.
        precomputed_dist (bool): If True, then input should be precomputed distance matrix,
            if False then input is array with features.
        sparse_dist (bool): If True, then use distance matrix in sparse format (zero elements
            are elements for which distance between points is greater than max_distance).
            If False, then consider distance matrix as ordinary numpy array.
        deterministic (bool): If True then take points one by one to get deterministic behavior.
            If False then select points at random, so results would be different for each run.
        verbose (bool): If True then output progress info, otherwise be silent.

    Attributes:
        labels_ (np.ndarray): Array with cluster labels after fitting model.
        n_clusters_ (int): Number of clusters after fitting model.
        centers_ (np.ndarray): Array with indexes of cluster centers.
        leaders_ (np.ndarray): Array with 1 for cluster centers and with 0 for all other points.
    """

    def __init__(self, max_radius: float = 0.1, change_leaders: bool = False,
                 metric: str = 'cosine', precomputed_dist: bool = False,
                 sparse_dist: bool = True, deterministic: bool = False,
                 verbose: bool = True):

        self.max_radius = max_radius
        self.change_leaders = change_leaders
        self.metric = metric
        self.precomputed_dist = precomputed_dist
        self.sparse_dist = sparse_dist
        self.deterministic = deterministic
        self.verbose = verbose

        self.max_distance = max_radius  # is needed for computation of sparse distance matrix
        self.labels_ = None
        self.n_clusters_ = None
        self.centers_ = None
        self.leaders_ = None

    def fit(self, X: Union[np.ndarray, csr_matrix]):
        """Fit clustering from features or distance matrix.

        Args:
            X (np.ndarray or scipy.sparse.csr_matrix): Array with features or
                precomputed distance matrix, could be in sparse matrix format.
        """

        dist_matrix = self._prepare_distance_matrix(X)

        # create arrays for labels, leaders and centers
        labels = np.empty(dist_matrix.shape[0])
        labels.fill(np.nan)
        leaders = np.zeros(dist_matrix.shape[0])
        centers = []

        # choose first point and assign label to it
        idx = 0 if self.deterministic else np.random.choice(range(len(labels)))
        labels[idx] = 0
        next_cluster = 1
        leaders[idx] = 1
        centers.append(idx)

        for _ in tqdm(range(len(labels)-1), desc='LeaderClustering fit', disable=not self.verbose):

            # choose next point
            indexes = np.where(np.isnan(labels))[0]
            idx = indexes[0] if self.deterministic else np.random.choice(indexes)
            # find indices of current leaders
            current_leaders_idx = np.where(leaders == 1)[0]
            current_leaders_labels = labels[current_leaders_idx]

            # find distances to current leaders
            leaders_dist = self._slice_distance_matrix(dist_matrix, idx, current_leaders_idx)

            if np.min(leaders_dist) <= self.max_radius:
                # assign cluster with nearest leader as label
                labels[idx] = current_leaders_labels[leaders_dist.argmin()]

                # change leader in cluster if there is better candidate for it
                if self.change_leaders:
                    cluster_idx = np.where(labels == labels[idx])[0]
                    dist_inside = dist_matrix[cluster_idx][:, cluster_idx].mean(axis=1)
                    min_idx = cluster_idx[dist_inside.argmin()]
                    nearest_leader_idx = current_leaders_idx[leaders_dist.argmin()]
                    if min_idx != nearest_leader_idx:
                        leaders[nearest_leader_idx] = 0
                        leaders[min_idx] = 1

            else:
                # assign new cluster label
                labels[idx] = next_cluster
                leaders[idx] = 1
                centers.append(idx)
                next_cluster += 1

        self.labels_ = labels.astype(int)
        self.n_clusters_ = int(labels.max() + 1)
        self.centers_ = np.array(centers)
        self.leaders_ = leaders.astype(int)



"""
Implementation of Quality threshold clustering.
"""

from typing import Union

import numpy as np
from scipy.sparse import csr_matrix, lil_matrix
from tqdm import tqdm

from .mixins import FitPredictMixin, DistanceMatrixMixin


class QTClustering(FitPredictMixin, DistanceMatrixMixin):
    """Quality threshold clustering.

    Args:
        max_radius (float): Maximum radius of cluster
            (maximum distance between center of cluster and all other points).
        min_cluster_size (int): Minimum size of clusters, stop iterations at this cluster size.
        metric (str): Distance metric.
            For sparse_dist=True possible options are in sklearn.neighbors.VALID_METRICS['brute'].
            For sparse_dist=False possible options are 'inner_product' or one of metrics
            available in scipy.spatial.distance.pdist. If 'inner_product' then use np.inner
            which is much faster than pdist. 'inner_product' could be used instead
            of cosine distance for normalized vectors.
        precomputed_dist (bool): If True, then input should be precomputed distance matrix,
            if False then input is array with features.
        sparse_dist (bool): If True, then use distance matrix in sparse format (zero elements
            are elements for which distance between points is greater than max_distance).
            If False, then consider distance matrix as ordinary numpy array.
        verbose (bool): If True then output progress info, otherwise be silent.

    Attributes:
        labels_ (np.ndarray): Array with cluster labels after fitting model.
        n_clusters_ (int): Number of clusters after fitting model.
        centers_ (np.ndarray): Array with indexes of cluster centers.
    """

    def __init__(self, max_radius: float = 0.1, min_cluster_size: int = 2,
                 metric: str = 'cosine', precomputed_dist: bool = False,
                 sparse_dist: bool = True, verbose: bool = True):

        self.max_radius = max_radius
        self.min_cluster_size = min_cluster_size
        self.metric = metric
        self.precomputed_dist = precomputed_dist
        self.sparse_dist = sparse_dist
        self.verbose = verbose

        self.max_distance = max_radius  # is needed for computation of sparse distance matrix
        self.labels_ = None
        self.n_clusters_ = None
        self.centers_ = None

    def fit(self, X: Union[np.ndarray, csr_matrix]):
        """Fit clustering from features or distance matrix.

        Args:
            X (np.ndarray or scipy.sparse.csr_matrix): Array with features or
                precomputed distance matrix, could be in sparse matrix format.
        """

        dist_matrix = self._prepare_distance_matrix(X)

        if self.sparse_dist:
            dist_mask = lil_matrix(dist_matrix)
            dist_mask[dist_mask > 0] = 1
            dist_mask.setdiag(1)
            labels, centers = self.fit_sparse(dist_mask)
        else:
            dist_mask = dist_matrix < self.max_radius
            np.fill_diagonal(dist_mask, True)
            labels, centers = self.fit_dense(dist_mask)

        self.labels_ = labels.astype(int)
        self.n_clusters_ = int(labels.max() + 1)
        self.centers_ = np.array(centers)

    def fit_dense(self, dist_mask: np.ndarray):
        """Fit clustering from distance matrix mask when it is dense matrix."""

        labels = np.empty(dist_mask.shape[0])
        labels.fill(np.nan)
        centers = []
        cluster_number = 0
        total_count = 0

        with tqdm(total=dist_mask.shape[0], disable=not self.verbose) as pbar:
            while dist_mask.any():

                # find size of candidate clusters for each point
                candidate_size = dist_mask.sum(axis=1)

                if np.max(candidate_size) < self.min_cluster_size:
                    labels[np.where(np.isnan(labels))] = -1
                    break

                # pick the biggest possible cluster from candidates
                center_idx = np.argmax(candidate_size)
                cluster_points_idx = np.where(dist_mask[center_idx])[0]
                # assign labels
                labels[cluster_points_idx] = cluster_number
                centers.append(center_idx)
                # remove labeled data from further calculations
                dist_mask[cluster_points_idx, :] = False
                dist_mask[:, cluster_points_idx] = False

                # finalize iteration
                cluster_number += 1
                size = np.max(candidate_size)
                total_count += size

                pbar.update(size)
                pbar.set_description(
                    f"QTClustering fit. Current cluster size {size}, total count {total_count}")

        return labels, centers

    def fit_sparse(self, dist_mask: csr_matrix):
        """Fit clustering from distance matrix mask when it is sparse matrix."""

        labels = np.empty(dist_mask.shape[0])
        labels.fill(np.nan)
        centers = []
        cluster_number = 0
        total_count = 0

        with tqdm(total=dist_mask.shape[0], disable=not self.verbose) as pbar:
            while dist_mask.sum() > 0:

                # find size of candidate clusters for each point
                candidate_size = dist_mask.sum(axis=1)

                if np.max(candidate_size) < self.min_cluster_size:
                    labels[np.where(np.isnan(labels))] = -1
                    break

                # pick the biggest possible cluster from candidates
                center_idx = np.argmax(candidate_size)
                cluster_points_idx = dist_mask[center_idx].nonzero()[1]
                # assign labels
                labels[cluster_points_idx] = cluster_number
                centers.append(center_idx)

                # remove labeled data from further calculations
                dist_mask[cluster_points_idx, :] = 0
                dist_mask[:, cluster_points_idx] = 0

                # finalize iteration
                cluster_number += 1
                size = np.max(candidate_size)
                total_count += size

                pbar.update(size)
                pbar.set_description(
                    f"QTClustering fit. Current cluster size {size}, total count {total_count}")

        return labels, centers
